//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     cf.h
// Purpose:    Conditional Facts
// Last Change:13.10.2001
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2001 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	CF_INCLUDED
#define CF_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef BOOL_INCLUDED
#include "bool.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef LLIST_INCLUDED
#include "llist.h"
#endif

#ifndef GLIT_INCLUDED
#include "glit.h"
#endif

#ifndef GNEG_INCLUDED
#include "gneg.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// CF_MAGIC: Magic number (identifies objects of cf_c for debugging).
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
static const long	CF_MAGIC =	0x43462000L;	// 'CF \0'
#endif

//-----------------------------------------------------------------------------
// CF_HD_MAGIC: Magic number (identifies objects of cf_hd_c for debugging).
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
static const long	CF_HD_MAGIC =	0x43464800L;	// 'CFH\0'
#endif

//-----------------------------------------------------------------------------
// CF_CN_MAGIC: Magic number (identifies objects of cf_cn_c for debugging).
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
static const long	CF_CN_MAGIC =	0x43464300L;	// 'CFC\0'
#endif

//=============================================================================
// Class for Conditional Facts:
//=============================================================================

class cf_c {

//-----------------------------------------------------------------------------
// Static Member Functions (Class Methods):
//-----------------------------------------------------------------------------

	public:

	// store: Store a conditional fact.
	static void store(glit_t *head, int head_size,
				gneg_t *cond, int cond_size);

	// store: Store a conditional fact.
	static void store(lit_t *head_lits, int head_size,
				lit_t *neg_lits, int neg_size);

	// Print all currently existing conditional facts:
	static void print_all(void);

	// diff_facts: Conditional facts inserted in the previous iteration.
	static inline cf_c* diff_facts(void)
		{ return(DiffFirst); }

	// old_facts: Conditional facts inserted before the previous iteration.
	static inline cf_c* old_facts(void)
		{ return(OldFirst); }

	// delete_obsolete: Delete the conditional facts that are now obsolete.
	static void delete_obsolete(void);
		// Warning: No pointer to a conditional fact should be held
		// outside this module when delete_obsolete is called
		// since it may destroy the conditional facts pointed to.

	// next_iter: Start the next iteration of hyperresolution.
	static void next_iter(void);

	// is_inconsistent: Does the set of conditional facts contain "FALSE"?
	static inline bool_t is_inconsistent(void)
		{ return(Inconsistent); }

	// reduce: Do positive and negative reduction.
	static void reduce(void);

//-----------------------------------------------------------------------------
// Member Functions (Object Methods):
//-----------------------------------------------------------------------------

	public:

	// Get pointer to head elements (list of facts):
	inline class cf_hd_c* get_head(void) const
		{ CHECK(valid(), "cf_c::get_head: Bad object");
		  return(Head); }

	// Get pointer to condition/body elements (list of default negations):
	inline class cf_cn_c* get_cond(void) const
		{ CHECK(valid(), "cf_c::get_cond: Bad object");
		  return(Cond); }

	// Was this conditional fact already discovered to be redundant?
	inline bool_t is_obsolete(void) const
		{ CHECK(valid(), "cf_c::is_obsolete: Bad object");
		  return(IsObsolete); }

	// Is this a conditional answer (not really part of residual program)?
	inline bool_t is_cond_answer(void) const
		{ CHECK(valid(), "cf_c::is_cond_answer: Bad object");
		  return(IsCondAnswer); }

	// Get pointer to next conditional fact (list of old/diff/new facts):
	inline cf_c* get_next(void) const
		{ CHECK(valid(), "cf_c::next: Bad object");
		  return(Next); }

	// Print this conditional fact:
	void print(void) const;

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity Check:
	public:
	bool_t	valid(void) const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "CF_MAGIC"
#endif

//-----------------------------------------------------------------------------
// Auxillary Functions (Private Member Functions):
//-----------------------------------------------------------------------------

	private:

	// Constructor:
	cf_c(bool_t unconditional, bool_t cond_answer);

	// Destructor:
	~cf_c(void);

	// Add a head literal:
	void add_head(class cf_hd_c *head_elem);

	// Add a condition element (default negation, negative body literal):
	void add_cond(class cf_cn_c *cond_elem);

	// Check for non-minimal conditional facts:
	bool_t redundant(long comparison, int length);

	// Set the Next pointer (list of old/diff/new facts):
	inline void set_next(cf_c *next)
		{ CHECK(valid(), "cf_c::set_next: Bad object");
		  // next is not yet completely initialized.
		  // The following check fails because next->Next == NIL,
		  // but none of the Last pointers points to it yet.
		  // CHECK(next == 0 || next->valid(),
		  //	  "cf_c::set_next: next is invalid");
		  // The statement following set_next() in the constructor
		  // solves the problem.
		  Next = next; }

	// Set the Prev pointer (list of old/diff/new facts):
	inline void set_prev(cf_c *prev)
		{ CHECK(valid(), "cf_c::set_prev: Bad object");
		  CHECK(prev == 0 || prev->valid(),
			  "cf_c::set_prev: prev is invalid");
		  Prev = prev; }

	// Get the next conditional fact on the list of obsolete ones:
	inline cf_c *get_next_obsolete(void)
		{ CHECK(valid(), "cf_c::get_next_obsolete: Bad object");
		  CHECK(IsObsolete, "cf_c::get_next_obsolete: Not obsolete");
		  return(ObsoleteNext); }

	// Mark this fact as obsolete:
	inline void mark_obsolete(void)
		{ CHECK(valid(), "cf_c::mark_obsolete: Bad object");
		  if(IsObsolete == BOOL_FALSE) {
			CHECK(ObsoleteNext == 0,
				"cf_c::mark_obsolete: inconsistent list");
			ObsoleteNext = Obsolete;
			Obsolete = this;
			IsObsolete = BOOL_TRUE;
		  }
		}

	// Get the Next pointer in the queue of facts for negative reduction:
	inline cf_c* get_neg_red_next(void) const
		{ CHECK(valid(), "cf_c::get_neg_red_next: Bad object");
		  CHECK(Cond == 0, "cf_c::get_neg_red_next: Has condition");
		  return(NegRedQueueNext); }

	// Set the Next pointer (queue of facts for negative reduction):
	inline void set_neg_red_next(cf_c *next)
		{ CHECK(valid(), "cf_c::set_neg_red_next: Bad object");
		  CHECK(Cond == 0, "cf_c::set_neg_red_next: Has condition");
		  CHECK(next == 0 || next->valid(),
			  "cf_c::set_neg_red_next: next is invalid");
		  CHECK(next == 0 || next->get_cond() == 0,
			  "cf_c::set_neg_red_next: next has condition");
		  NegRedQueueNext = next; }

	// Set the Prev pointer (queue of facts for negative reduction):
	inline void set_neg_red_prev(cf_c *prev)
		{ CHECK(valid(), "cf_c::set_neg_red_prev: Bad object");
		  CHECK(Cond == 0, "cf_c::set_neg_red_prev: Has condition");
		  CHECK(prev == 0 || prev->valid(),
			  "cf_c::set_neg_red_prev: prev is invalid");
		  CHECK(prev == 0 || prev->get_cond() == 0,
			  "cf_c::set_neg_red_prev: prev has condition");
		  NegRedQueuePrev = prev; }

	// neg_reduce: Do one negative reduction step.
	void neg_reduce(void);

	// neg_reduce_test: Check overlap of condition with disjunctive fact:
	void neg_reduce_test(int length);

//-----------------------------------------------------------------------------
// Auxillary Class Functions (Private Static Member Functions):
//-----------------------------------------------------------------------------

	// pos_reduce: Do positive reduction.
	static void pos_reduce(glit_t lit);

//-----------------------------------------------------------------------------
// Static Data Members (Class Attributes):
//-----------------------------------------------------------------------------

	private:

	// Unique number of current comparison:
	static long ComparID;

	// First conditional fact inserted in this hyperresolution iteration:
	static cf_c* NewFirst;

	// Last conditional fact inserted in this hyperresolution iteration:
	static cf_c* NewLast;

	// First conditional fact inserted in the previous iteration:
	static cf_c* DiffFirst;

	// Last conditional fact inserted in the previous iteration:
	static cf_c* DiffLast;

	// First conditional fact inserted before the previous iteration:
	static cf_c* OldFirst;

	// Last conditional fact inserted before the previous iteration:
	static cf_c* OldLast;

	// Conditional Facts that became obsolete due to a newly inserted one:
	static cf_c* Obsolete;

	// The empty clause ("FALSE") was derived:
	static bool_t Inconsistent;

	// Unconditional disjunctions not yet processed for negative reduction:
	static cf_c* NegRedQueue;

	// End of queue for negative reduction:
	static cf_c* NegRedQueueLast;

//-----------------------------------------------------------------------------
// Data Members (Object Attributes):
//-----------------------------------------------------------------------------

	private:

	// Link to next conditional fact:
	cf_c *Next;

	// Link to previous conditional fact:
	cf_c *Prev;

	// Head of this conditional fact:
	class cf_hd_c *Head;

	// Last element of the head list so far:
	class cf_hd_c *LastHead;

	// Condition (Body) of this conditional fact:
	class cf_cn_c *Cond;

	// Last element of the condition list so far:
	class cf_cn_c *LastCond;

	// Total length of the clause (number of literals in head and cond).
	int Length;

	// Overlap counter (number of common literals for current comparison):
	int Overlap;

	// Number of the comparison when Overlap was last set:
	long LastSet;

	// Is this fact redundant/obsolete (should be deleted soon)?
	bool_t IsObsolete;

	// Is this a conditional answer (not really part of residual program)?
	bool_t IsCondAnswer;

	// Link for the list of obsolete conditional facts (too be deleted):
	class cf_c *ObsoleteNext;

	// Link to next disjunctive fact in queue for negative reduction:
	cf_c *NegRedQueueNext;

	// Link to previous disjunctive fact in queue for negative reduction:
	cf_c *NegRedQueuePrev;

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Declaration of Pointer Type:
//-----------------------------------------------------------------------------

typedef cf_c *cf_t;

//=============================================================================
// Class for Elements of the Head of a Conditional Fact:
//=============================================================================

class cf_hd_c {

	public:

//-----------------------------------------------------------------------------
// Methods:
//-----------------------------------------------------------------------------

	// Constructor:
	cf_hd_c(glit_t fact, cf_c *cf);

	// Destructor:
	~cf_hd_c(void);

	// Return the conditional fact to which this head element belongs:
	inline cf_c *get_cf(void) const
		{ CHECK(valid(), "cf_hd_c::get_cf: Bad object");
		  return(Cf); }

	// Return the fact (ground literal) of this head element:
	inline glit_t get_fact(void) const
		{ CHECK(valid(), "cf_hd_c::get_fact: Bad object");
		  return(Fact); }

	// Return the next occurrence of the same fact in a head:
	inline cf_hd_c *get_next_fact_occur(void) const
		{ CHECK(valid(), "cf_hd_c::get_next_fact_occur: Bad object");
		  return(NextFactOccur); }

	// Set link to the next occurrence of the same fact in a head:
	inline void set_next_fact_occur(cf_hd_c *next)
		{ CHECK(valid(), "cf_hd_c::set_next_fact_occur: Bad object");
		  CHECK(next == 0 || next->valid(),
			"cf_hd_c::set_next_fact_occur: next is invalid");
		  NextFactOccur = next; }

	// Set link to the previous occurrence of the same fact in a head:
	inline void set_prev_fact_occur(cf_hd_c *prev)
		{ CHECK(valid(), "cf_hd_c::set_prev_fact_occur: Bad object");
		  CHECK(prev == 0 || prev->valid(),
			"cf_hd_c::set_prev_fact_occur: prev is invalid");
		  PrevFactOccur = prev; }

	// Return the next element of the head of the same conditional fact:
	inline cf_hd_c *get_next_head_elem(void) const
		{ CHECK(valid(), "cf_hd_c::get_next_head_elem: Bad object");
		  return(NextHeadElem); }

	// Append another element to the list of head literals of a cond fact:
	void append(cf_hd_c *hd)
		{ CHECK(valid(), "cf_hd_c::append: Bad object");
		  CHECK(hd != 0, "cf_hd_c::append: hd is NIL");
		  CHECK(hd->valid(), "cf_hd_c::append: hd is invalid");
		  CHECK(NextHeadElem == 0, "cf_hd_c::append: Next already set");
		  NextHeadElem = hd; }
		  
//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity Check:
	public:
	bool_t	valid(void) const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "CF_HD_MAGIC"
#endif

//-----------------------------------------------------------------------------
// Data Members (Object Attributes):
//-----------------------------------------------------------------------------

	private:

	// Reference back to Conditional Fact:
	cf_c	*Cf;

	// Fact in this head element:
	glit_t	Fact;

	// Pointer to next element of this head:
	cf_hd_c *NextHeadElem;

	// Link to next head element with the same fact (different rule):
	cf_hd_c *NextFactOccur;

	// Link to previous head element with the same fact (different rule):
	cf_hd_c *PrevFactOccur;

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Declaration of Pointer Type:
//-----------------------------------------------------------------------------

typedef cf_hd_c *cf_hd_t;

//=============================================================================
// Class for Elements of the Condition (Body) of a Conditional Fact:
//=============================================================================

class cf_cn_c {

	public:

//-----------------------------------------------------------------------------
// Methods:
//-----------------------------------------------------------------------------

	// Constructor:
	cf_cn_c(gneg_t neg, cf_c *cf);

	// Destructor:
	~cf_cn_c(void);

	// Return the conditional fact to which this head element belongs:
	inline cf_c *get_cf(void) const
		{ CHECK(valid(), "cf_cn_c::get_cf: Bad object");
		  return(Cf); }

	// Return the ground default negation of this condition element:
	inline gneg_t get_neg(void) const
		{ CHECK(valid(), "cf_cn_c::get_neg: Bad object");
		  return(Neg); }

	// Return the next occurrence of the same default negation in a cond:
	inline cf_cn_c *get_next_neg_occur(void) const
		{ CHECK(valid(), "cf_cn_c::get_next_neg_occur: Bad object");
		  return(NextNegOccur); }

	// Set link to next occurrence of the same default negation in a cond:
	inline void set_next_neg_occur(cf_cn_c *next)
		{ CHECK(valid(), "cf_cn_c::set_next_neg_occur: Bad object");
		  CHECK(next == 0 || next->valid(),
			"cf_cn_c::set_next_neg_occur: next is invalid");
		  NextNegOccur = next; }

	// Set link to previous occurrence of the default negation in a cond:
	inline void set_prev_neg_occur(cf_cn_c *prev)
		{ CHECK(valid(), "cf_cn_c::set_prev_neg_occur: Bad object");
		  CHECK(prev == 0 || prev->valid(),
			"cf_cn_c::set_prev_neg_occur: prev is invalid");
		  PrevNegOccur = prev; }

	// Return the next element of the condition of the same cond fact:
	inline cf_cn_c *get_next_cond_elem(void) const
		{ CHECK(valid(), "cf_cn_c::get_next_cond_elem: Bad object");
		  return(NextCondElem); }

	// Append another element to the list of condtions of a cond fact:
	void append(cf_cn_c *cond_elem)
		{ CHECK(valid(), "cf_cn_c::append: Bad object");
		  CHECK(cond_elem != 0, "cf_cn_c::append: cond_elem is NIL");
		  CHECK(cond_elem->valid(),
			"cf_cn_c::append: cond_elem is invalid");
		  CHECK(NextCondElem == 0,
			"cf_cn_c::append: Next already set");
		  NextCondElem = cond_elem; }
		  
//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity Check:
	public:
	bool_t	valid(void) const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "CF_BD_MAGIC"
#endif

//-----------------------------------------------------------------------------
// Data Members (Object Attributes):
//-----------------------------------------------------------------------------

	private:

	// Reference back to Conditional Fact:
	cf_c	*Cf;

	// Default negation in this condition element:
	gneg_t	Neg;

	// Pointer to next element of this condition:
	cf_cn_c *NextCondElem;

	// Link to next occurrence of the same default negation in a condition:
	cf_cn_c *NextNegOccur;

	// Link to previous occurrence of the same default negation:
	cf_cn_c *PrevNegOccur;

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Declaration of Pointer Type:
//-----------------------------------------------------------------------------

typedef cf_cn_c *cf_cn_t;

//=============================================================================
// End of Include File:
//=============================================================================

#endif

